from __future__ import annotations
from fastgpx.fastgpx import Bounds
from fastgpx.fastgpx import Gpx
from fastgpx.fastgpx import LatLong
from fastgpx.fastgpx import Segment
from fastgpx.fastgpx import TimeBounds
from fastgpx.fastgpx import Track
from fastgpx.fastgpx import geo
from fastgpx.fastgpx import parse
from fastgpx.fastgpx import polyline
from . import fastgpx
__all__: list = ['Bounds', 'Gpx', 'LatLong', 'Segment', 'TimeBounds', 'Track', 'geo', 'parse', 'polyline']
